
class ProxyLogger(object):
    """
    This class provides a wrapper for the standard Python logging
    facilities.  Handlers do not store their messages.  Their messages
    flush immediately to whereever they are intended to go.

    ProxyLogger proxies those received messages along, but also stores them,
    so that they can be retrieved and enumerated later.
    """

    def __init__(self, logger):
        self._messages = []
        self.logger = logger

    def get_messages(self):
        return self._messages

    def flush(self):
        """Clear all messages and return them afterwards."""
        messages = self._messages
        self._messages = []
        return messages

    def warning(self, msg):
        self._messages.append(("WARNING", msg))
        self.logger.warning(msg)

    def info(self, msg):
        self._messages.append(("INFO", msg))
        self.logger.info(msg)

    def error(self, msg):
        self._messages.append(("ERROR", msg))
        self.logger.error(msg)

    def debug(self, msg):
        self._messages.append(("DEBUG", msg))
        self.logger.debug(msg)

    def has_errors(self):
        return any(et == 'ERROR' for et, _ in self._messages)

    def has_warnings(self):
        return any(et == 'WARNING' for et, _ in self._messages)



class Q(object):
    """
    Query class used to build arbitrarily complex query filters.

    Q objects are strung together and then compiled once we are told what the
    concrete LDAPNode class is going to be.

    The key to the Q class is that it appends conditions when of the same type
    without creating another level of hierarchy, and it spawns child levels
    only when the operations (And, Or) switches.
    """

    def __init__(self, **conditions):
        self.conditions = conditions.items()

    def compile(self, cls):
        immediate_filter = ''
        for cond_k, cond_v in self.conditions:
            try:
                attrname = cls._fields.get(cond_k).ldap
            except AttributeError:
                raise AttributeError(
                    f'{cond_k} is not a valid field on {cls.__name__}: '
                    f'Expected: {list(cls._fields.keys())}')
            immediate_filter += f'({attrname}={cond_v})'

        if len(self.conditions) > 1:
            immediate_filter = f'(&{immediate_filter})'

        return immediate_filter

    def check_type_compat(self, other):
        if not isinstance(other, Q):
            raise TypeError(f'not of type Q: {other}')

    def __or__(self, other):
        self.check_type_compat(other)
        if isinstance(other, Or):
            other.ops.insert(0, self)
            return other
        else:
            return Or([self, other])

    def __and__(self, other):
        self.check_type_compat(other)
        if isinstance(other, And):
            other.ops.insert(0, self)
            return other
        else:
            return And([self, other])


class And(Q):

    def __init__(self, ops):
        self.ops = ops

    def compile(self, cls):
        f = ''
        for op in self.ops:
            f += op.compile(cls)
        return f'(&{f})'

    def __and__(self, other):
        if type(other) is Q:  # And(...) & Q(...)
            # we repack Q(...) objects into one Q() object for each condition.
            # This is so that:
            #   (Q(lastname='Lennon') & Q(lastname='McCartney')) & \
            #       Q(firstname='Ringo', lastname='Starr')
            # Compiles to: (&(sn=Lennon)(sn=McCartney)(givenName=Ringo)(sn=Starr))
            # And not:     (&(sn=Lennon)(sn=McCartney)(&(givenName=Ringo)(sn=Starr)))
            self.ops.extend([Q(**{k: v}) for k, v in other.conditions])
            return self
        elif type(other) is And:  # And(...) & And(...)
            self.ops.extend(other.ops)
            return self
        else:  # And(...) & Or(...)
            return super().__and__(other)


class Or(Q):

    def __init__(self, ops):
        self.ops = ops

    def compile(self, cls):
        f = ''
        for op in self.ops:
            f += op.compile(cls)
        return f'(|{f})'

    def __or__(self, other):
        if type(other) is Q:
            self.ops.append(other)
            return self
        elif type(other) is Or:
            self.ops.extend(other.ops)
            return self
        else:
            return super().__or__(other)



# -*- coding: utf-8 -*-

import pytest

from .test_ldapnode import get_person

from ldapper.exceptions import InvalidDN
from ldapper.utils import (
    bolded,
    dn_attribute,
    strip_dn_path,
    middle_dn,
    get_attr,
    get_attrlist,
    inflect_given_cardinality,
    build_ldap_filter,
    list_items_to_sentence,
    remove_empty_strings,
    stringify,
    to_list,
    print_word_list,
    validate_dn,
)


class TestUtils(object):

    def test_bolded(self):
        assert bolded('foobar') == '\033[1mfoobar\033[0m'

    def test_dn_attribute(self):
        assert dn_attribute('cn=foo,dc=bar', 'cn') == 'foo'
        assert dn_attribute('cn=foo,cn=bar,dc=ba', 'cn') == 'foo'
        assert dn_attribute('cn=foo,dc=bar', 'uid') is None

    def test_strip_dn_path(self):
        dn = 'cn=foo,ou=groups,dc=example'
        left, right = 'cn=foo,', ',dc=example'
        assert strip_dn_path(dn) == dn
        assert strip_dn_path(dn, left=left) == 'ou=groups,dc=example'
        assert strip_dn_path(dn, right=right) == 'cn=foo,ou=groups'
        assert strip_dn_path(dn, left=left, right=right) == 'ou=groups'

    def test_middle_dn(self):
        dn = 'cn=foo,ou=middle-mgrs,ou=groups,dc=example'
        right = ',dc=example'
        assert middle_dn(dn, right) == 'ou=middle-mgrs,ou=groups'
        assert middle_dn('cn=foo', right='aaa') is None

    def test_get_attr(self):
        assert get_attr(entry={}, attr='name') is None
        assert get_attr(entry={'name': ['Liam', 'Mon']}, attr='name') == 'Liam'

    def test_get_attrlist(self):
        assert get_attrlist(entry={}, attr='name') == []
        assert get_attrlist(entry={'name': ['L', 'M']}, attr='name') == ['L', 'M']

    def test_inflect_given_cardinality(self):
        assert inflect_given_cardinality('bird', 0) == 'birds'
        assert inflect_given_cardinality('bird', 1) == 'bird'
        assert inflect_given_cardinality('bird', 2) == 'birds'

    @pytest.mark.parametrize("kwargs,expected", [
        ({'op': '&', 'attrname': 'objClass'}, '(&(objClass=foo)(objClass=bar))'),
        ({'op': '|', 'attrname': 'fooClass'}, '(|(fooClass=foo)(fooClass=bar))'),
    ])
    def test_build_ldap_filter(self, kwargs, expected):
        assert build_ldap_filter(items=['foo', 'bar'], **kwargs) == expected

    def test_build_ldap_filter_no_items(self):
        with pytest.raises(ValueError):
            assert build_ldap_filter(op='&', attrname='objectClass', items=None) == 'foo'
        with pytest.raises(ValueError):
            assert build_ldap_filter(op='&', attrname='objectClass', items=[]) == 'foo'

    def test_list_items_to_sentence(self):
        words = ['one', 'two', 'three']
        assert list_items_to_sentence(words[:1]) == 'one'
        assert list_items_to_sentence(words[:2]) == 'one and two'
        assert list_items_to_sentence(words[:3]) == 'one, two, and three'
        assert list_items_to_sentence([]) is None

    def test_print_word_list(self):
        words = [
            'one', 'two', 'three', 'four', 'five', 600, 'seven', 'eight',
            'nine', 'ten', 'eleven', 'twelve', 'thirteen', 'foobarr', 'fourteen',
        ]
        expexcted = '''\
 one two three four five 600 seven eight nine ten eleven twelve thirteen
 foobarr fourteen'''
        assert print_word_list(words) == expexcted

    def test_remove_empty_strings(self):
        assert remove_empty_strings('foo') == 'foo'
        assert remove_empty_strings(['foo', 'bar', '', None]) == ['foo', 'bar']
        assert remove_empty_strings(False) is False
        assert remove_empty_strings('') is None

    def test_stringify(self):
        p = get_person()
        assert {'name': 'foo'} == stringify({'name': 'foo'})
        assert {'name': 'foo', 'user': p.dnattr()} == \
            stringify({'name': 'foo', 'user': p})

    def test_to_list(self):
        assert [1, 2, 3] == to_list([1, 2, 3])
        assert ["foo"] == to_list("foo")
        assert [u"foo"] == to_list(u"foo")
        assert [] == to_list(3)

    def test_validate_dn(self):
        with pytest.raises(InvalidDN):
            validate_dn('foo')
        with pytest.raises(InvalidDN):
            validate_dn('cn=foo,')
        validate_dn('cn=foo,ou=bar')



# -*- coding: utf-8 -*-

import ldap
import pytest

from .utils import Connection
from .test_ldapnode import get_person


class TestConnection:

    def test_connection_init(self):
        conn = Connection(logindn=None, password=None, basedn='foo')
        assert conn.basedn == 'foo'

        uri = 'ldap://127.0.0.1:389'
        conn = Connection(logindn=None, password=None, uri=uri)
        assert conn.uri == uri

    def test_connnection_connect_failed_attempt(self, caplog):
        conn = Connection.connect(logindn='foo', password='wrong',
                                  retries=1, uri='ldap://localhost:389')
        assert conn is None

        assert 'login attempt 1 failed' in caplog.text

    def test_connection_is_anonymous(self):
        conn = Connection.connect_anon()
        assert conn.is_anonymous()
        assert conn.whoami() is None
        assert conn.whoami_short() is None

    def test_connection_fully_qualify_dn(self):
        conn = Connection.connect_anon()

        uid = 'user'
        dn = 'uid=user,ou=people,dc=acme,dc=org'

        # a shortname is transformed into a fully qualified dn
        assert conn._fully_qualify_dn(logindn=uid) == dn

        # an already fully qualified dn is left unchanged
        assert conn._fully_qualify_dn(logindn=dn) == dn

    def test_connection_str(self, connection):
        expected = 'ldap://localhost:389 as cn=admin,dc=acme,dc=org'
        assert str(connection) == expected

    def test_connection_whoami(self, connection):
        assert connection.whoami() == 'cn=admin,dc=acme,dc=org'
        assert connection.whoami_short() == 'admin'

    def test_connection_exists(self, connection):
        # happy path
        assert connection.exists('cn=admin,dc=acme,dc=org')

        # exists() does not try to catch this error
        with pytest.raises(ldap.INVALID_DN_SYNTAX):
            connection.exists('cnnn=foo,dc=acme,dc=org')

        assert connection.exists('cn=liam,dc=acme,dc=org') is False
        assert connection.exists('cn=🎂,dc=acme,dc=org') is False

    def test_connection_ldif(self, connection):
        p1 = get_person()
        p1.uid = 'bob'
        p2 = get_person()
        p2.uid = 'alice'
        p1.save()
        p2.save()

        output = connection.ldif('uid=*')
        assert 'DN: uid=bob,ou=people,dc=acme,dc=org' in output
        assert 'DN: uid=alice,ou=people,dc=acme,dc=org' in output

        output = connection.ldif('uid=bob')
        assert 'DN: uid=bob,ou=people,dc=acme,dc=org' in output
        assert 'DN: uid=alice,ou=people,dc=acme,dc=org' not in output

        p1.delete()
        p2.delete()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

from ldapper.logging import ProxyLogger


class TestProxyLogger:

    def test_logging_init(self):
        logger = ProxyLogger(logger=logging.getLogger('test'))
        assert not logger.has_errors()
        assert not logger.has_warnings()
        assert logger.get_messages() == []

    def test_logging_warning(self):
        logger = ProxyLogger(logger=logging.getLogger('test'))
        logger.warning('this is a test')
        assert logger.has_warnings()
        assert not logger.has_errors()

        assert logger.get_messages() == [("WARNING", 'this is a test')]
        # does not flush until we ask it to
        assert logger.get_messages() == [("WARNING", 'this is a test')]

        # we flush and get our messages
        assert logger.flush() == [("WARNING", 'this is a test')]
        # but now it should be empty
        assert logger.flush() == []
        assert logger.get_messages() == []

    def test_logging_info(self):
        logger = ProxyLogger(logger=logging.getLogger('test'))
        logger.info('this is an info msg')
        assert logger.get_messages() == [("INFO", 'this is an info msg')]
        assert not logger.has_warnings()
        assert not logger.has_errors()

    def test_logging_debug(self):
        logger = ProxyLogger(logger=logging.getLogger('test'))
        logger.debug('this is a debug msg')
        assert logger.get_messages() == [("DEBUG", 'this is a debug msg')]

    def test_logging_error(self):
        logger = ProxyLogger(logger=logging.getLogger('test'))
        logger.error('this is an error msg')
        assert logger.get_messages() == [("ERROR", 'this is an error msg')]
        assert logger.has_errors()
        assert not logger.has_warnings()



from ldapper.utils import (
    ad_date_parse,
    dn_attribute,
    get_attr,
    get_attrlist,
    to_list,
)


class Field(object):

    """
    A representation of an LDAP attribute.

    A ``Field`` knows how to take a value from LDAP and convert it to a Python
    object.  It know how to go the other way, too: taking a Python value
    and converting it to a representation suitable for LDAP modifications.
    """

    derived = False
    """
    Is the ``Field`` derived/inferred/computed?

    When ``derived=True``, the ``Field``:

      * is derived from something like the DN usually
      * is not a concrete attribute in ldap and thus cannot
        be saved to the LDAP
      * should not appear in diff() results
    """

    # System attributes are properties managed by the LDAP that cannot be
    # set by the user.  Things like createTimestamp, modifiedTimestamp
    system = False

    def __init__(self, ldap, optional=False, readonly=False, printable=True,
                 primary=False):
        """
        Initialize a Field.

        Arguments:
            ldap - the name of the LDAP attribute
            optional - controls whether the field is mandatory during
                       object creation. Defaults to False.
            readonly - The field cannot be modified after creation.
            printable - True if the field should show up in pretty_print()
            primary - True if this is the field that primarily identifies
                    the object.  Defaults to False.  Can only be one per class.
        """
        self.ldap = ldap
        self.optional = optional
        self.readonly = readonly
        self.printable = printable
        self.primary = primary

    def default_value(self):
        """The value used for the field if none is provided."""
        return None

    def coerce_for_python(self, value):
        """
        Returns a transformed value for the ``value`` provided.

        This will get invoked both when loading a value for this Field out
        of the LDAP or when creating a new object.

        The default implementation returns the value unchanged.
        """
        return value


class IntegerField(Field):

    def populate(self, dn, entry):
        val = get_attr(entry, self.ldap)
        if val:
            return val.decode('utf-8')
        else:
            return val

    def coerce_for_python(self, value):
        if value is None and self.optional:
            return None
        try:
            return int(value)
        except ValueError:
            msg = "%s must be an int: got %s" % (self.ldap, value)
            raise ValueError(msg)
        except TypeError:
            msg = "value for %s cannot be converted to an integer" % self.ldap
            raise ValueError(msg)

    def sanitize_for_ldap(self, val):
        if val is None:
            return None
        else:
            return str(val).encode('utf-8')


class StringField(Field):

    def populate(self, dn, entry):
        val = get_attr(entry, self.ldap)
        if val:
            return val.decode('utf-8')
        else:
            return val

    def coerce_for_python(self, value):
        if type(value) == int:
            return str(value)
        return value

    def sanitize_for_ldap(self, val):
        if val:
            if not isinstance(val, str):
                val = str(val)
            return val.encode('utf-8')
        else:
            return None


class ListField(Field):

    """Lists can only hold Strings at this time."""

    def default_value(self):
        return []

    def populate(self, dn, entry):
        vals = get_attrlist(entry, self.ldap)
        if vals:
            return [v.decode('utf-8') for v in vals]
        else:
            return vals

    def coerce_for_python(self, value):
        return to_list(value)

    def sanitize_for_ldap(self, val):
        return [v.encode('utf-8') for v in val]


class DNPartField(Field):
    """
    A field that comes from a portion of the object's DN.

    e.g. If we had DN: cn=foo,device=device1,ou=devices

        Then we could specify `device` as the part, and get back `device1`
        as the field.

    There is no ability to make a DNPartField optional, since it is
    a part of the DN.
    """

    derived = True

    def populate(self, dn, entry):
        return dn_attribute(dn, self.ldap)

    def coerce_for_python(self, value):
        primary = getattr(value, 'primary', None)
        if primary:
            return getattr(value, primary)
        else:
            return value

    # this field type should never be saved to LDAP, so there is no
    # sanitize_for_ldap() method defined.
    def sanitize_for_ldap(self, val):
        raise RuntimeError('DNPartField.sanitize_for_ldap() should never get called.')


class BinaryField(Field):

    def populate(self, dn, entry):
        return get_attr(entry, self.ldap)

    def sanitize_for_ldap(self, val):
        return val


class SystemField(Field):
    """
    Meant for ldap attributes that are system-managed by the LDAP server.
    """

    system = True

    def sanitize_for_ldap(self, val):
        raise Exception('This Field may not be written to the LDAP')


class Win32TimestampField(StringField):
    """
    The 18-digit Active Directory timestamps, also named 'Windows NT time format'
    """

    def coerce_for_python(self, value):
        if value:
            return ad_date_parse(value)



import re
import ldap
import datetime as dt

from inflection import singularize, pluralize
from ldapper.exceptions import InvalidDN


WIN32_EPOCH = dt.datetime(1601, 1, 1)


def ad_date_parse(ts):
    """Parse an ActiveDirectory timestamp and return a datetime object"""
    return WIN32_EPOCH + dt.timedelta(seconds=(int(ts) / 10000000))


def bolded(val):
    """return a bolded version of the string passed in"""
    return '%s%s%s' % ('\033[1m', val, '\033[0m')


def dn_attribute(dn, attr):
    """Given a full DN return the value of the attribute given"""
    for rdn in dn.split(','):
        if rdn.startswith('%s=' % attr):
            return rdn.split('=', 1)[1]


def strip_dn_path(dn, left=None, right=None):
    """Strip a given string on the right or left or both
       This is case insenstive"""
    dn_path = dn
    if right is not None:
        rpattern = re.compile(right + '$', re.IGNORECASE)
        dn_path = rpattern.sub('', dn_path)
    if left is not None:
        lpattern = re.compile('^' + left, re.IGNORECASE)
        dn_path = lpattern.sub('', dn_path)
    return dn_path


def middle_dn(dn, right):
    """Strip the first DN component and the right and return what's left."""
    try:
        first, rest = dn.split(',', 1)
    except ValueError:
        return None
    return strip_dn_path(rest, right=right)


def get_attr(entry, attr):
    """
    Return the first value for an attribute in the given entry.

    If the attribute does not exist in the entry, return `None`.
    """
    a = entry.get(attr)
    if a is not None:
        return a[0]
    else:
        return None


def get_attrlist(entry, attr):
    """
    Returns the whole list of values for an attribute in the given entry.

    If the attribute does not exist in the entry, return `None`.
    """
    a = entry.get(attr)
    if a is not None:
        return a
    else:
        return []


def inflect_given_cardinality(word, num_items):
    """
    Return the singular form of the word if `num_items` is 1.  Otherwise,
    return the plural form of the word.
    """
    if num_items == 1:
        return singularize(word)
    else:
        return pluralize(word)


def build_ldap_filter(op='&', attrname='objectClass', items=None):
    """Return an LDAP search filter string.

    Parameters
    ----------
    op : str
        The search filter operator to use.
        Commonly one of the following based on the LDAP server dialect:

        +------------------+------------------------+
        | Logical Operator | Description            |
        +==================+========================+
        | &                | AND                    |
        +------------------+------------------------+
        | |                | OR                     |
        +------------------+------------------------+
        | !                | NOT                    |
        +------------------+------------------------+
        | =                | Equal to               |
        +------------------+------------------------+
        | ~=               | Approximately equal to |
        +------------------+------------------------+

    attrname : str
        The attribute to search on.  Forms the left-hand part
        of the filter subquery expressions.
    items : list
        The items to create subfilters over.

    Returns
    -------
    str
        the constructed LDAP filter.

    Raises
    ------
    ValueError
        if there are no ``items``.
    """
    if not items:
        raise ValueError("items list must exist")
    inside = ''.join(['(%s=%s)' % (attrname, item) for item in items])
    return '(%s%s)' % (op, inside)


def list_items_to_sentence(items):
    """
    Given a list, return a sentence containting the items separated by
    commas and the last element separated by "and" if there are more than
    two items.
    """
    if len(items) == 1:
        return str(items[0])
    elif len(items) == 2:
        return '%s and %s' % (items[0], items[1])
    elif len(items) > 1:
        return '%s, and %s' % (', '.join(map(str, items[:-1])), items[-1])


def print_word_list(words, line_length=79):
    """Given a list, print it in a condensed form separated by spaces"""
    chars_left_in_line = line_length
    returning = ''

    for word in words:
        if not isinstance(word, str):
            word = str(word)
        if chars_left_in_line < len(word) + 1:
            returning += '\n'
            chars_left_in_line = line_length
        returning += ' %s' % (word)
        chars_left_in_line -= (len(word) + 1)

    return returning


def remove_empty_strings(val):
    """
    Remove emptry strings from lists and return None if given an empty
    string.  This method is here so that we can tell if two attribute values
    are superficially different or not.

    For example, [] and [''] should not be considered different.  '' and None
    should also not be considered different.
    """
    if type(val) is list:
        return [x for x in val if x]
    elif isinstance(val, str):
        if len(val) == 0:
            return None
        else:
            return val
    else:
        return val


def stringify(args):
    """
    Given a hash, turn values that are LDAPNode objects into string
    representations based off the primary dn attribute.
    """
    returning = {}
    for attr in args:
        primary = getattr(args[attr], 'primary', None)
        if primary:
            returning[attr] = getattr(args[attr], primary)
        else:
            returning[attr] = args[attr]
    return returning


def to_list(possible_lst):
    """
    Coerce argument to a list by all means.

    Lists are returned unchanged, strings return a list with a single element,
    and otherwise an empty list is returned.
    """
    if isinstance(possible_lst, list):
        return possible_lst
    elif isinstance(possible_lst, str):
        return [possible_lst]
    else:
        return []


def validate_dn(dn, class_name=None, object_name=None):
    try:
        ldap.dn.str2dn(dn)
    except ldap.DECODING_ERROR:
        raise InvalidDN(class_name, object_name, dn)



# -*- coding: utf-8 -*-

import pytest

from ldapper.fields import (
    BinaryField,
    DNPartField,
    IntegerField,
    ListField,
    StringField,
)

from .utils import MyLDAPNode


class TestStringField:

    def test_populate(self):
        dn = 'cn=foo'
        field = StringField('name')

        # test the happy path
        assert field.populate(dn, {'name': [b'liam']}) == 'liam'

        # test if there is no entry from ldap for this field
        assert field.populate(dn, {}) is None

    def test_coerce_for_python(self):
        field = StringField('name')
        assert field.coerce_for_python(11) == '11'
        assert field.coerce_for_python('foobar') == 'foobar'

    def test_sanitize_for_ldap(self):
        field = StringField('name')
        assert field.sanitize_for_ldap(None) is None
        assert field.sanitize_for_ldap('foobar') == b'foobar'
        assert field.sanitize_for_ldap(12) == b'12'


class TestListField:

    def test_default_value(self):
        assert ListField('cns').default_value() == []

    def test_populate(self):
        dn = 'cn=foo'
        field = ListField('cn')
        assert field.populate(dn, {'cn': [b'liam', b'william']}) == ['liam', 'william']
        assert field.populate(dn, {}) == []

    def test_coerce_for_python(self):
        field = ListField('cn')
        assert field.coerce_for_python(None) == []
        assert field.coerce_for_python('foo') == ['foo']
        assert field.coerce_for_python(['foo']) == ['foo']

    def test_sanitize_for_ldap(self):
        field = ListField('cn')
        assert field.sanitize_for_ldap(['foo', 'bar']) == [b'foo', b'bar']


class TestIntegerField:

    def test_populate(self):
        field = IntegerField('id')
        assert field.populate('cn=foo', {'id': [b'12']}) == '12'
        assert field.populate('cn=foo', {}) is None

    def test_coerce_for_python_required(self):
        field = IntegerField('id')
        assert field.coerce_for_python('11') == 11
        with pytest.raises(ValueError):
            field.coerce_for_python('foo')
        with pytest.raises(ValueError):
            field.coerce_for_python(None)

    def test_coerce_for_python_optional(self):
        field = IntegerField('id', optional=True)
        assert field.coerce_for_python('11') == 11

        # None is ok, but values that truly don't parse will still fail
        assert field.coerce_for_python(None) is None
        with pytest.raises(ValueError):
            field.coerce_for_python('foo')

    def test_sanitize_for_ldap(self):
        field = IntegerField('id')
        assert field.sanitize_for_ldap(11) == b'11'
        assert field.sanitize_for_ldap(None) is None


class TestDNPartField:

    def test_dn_part_field_populate(self):
        field = DNPartField('name')
        assert field.populate('name=foo,dc=acme,dc=org', {}) == 'foo'

    def test_dn_part_field_coerce_for_python(self):
        class Foo(MyLDAPNode):
            uid = StringField('uid', primary=True)

        field = DNPartField('name')

        # make sure that if a DNPartField is passed an LDAPNode that it
        # uses the object's primary field value
        liam = Foo(uid='liam')
        assert field.coerce_for_python(liam) == 'liam'

        # if this is anything else, then there is no coercion
        assert field.coerce_for_python('byron') == 'byron'

    def test_dn_part_field_sanitize_for_ldap(self):
        field = DNPartField('name')
        with pytest.raises(RuntimeError):
            field.sanitize_for_ldap('foo')


class TestBinaryField:

    def test_binary_field_populate(self):
        entry = {'photo': [b'01010101']}
        photo = BinaryField('photo')
        # don't try to make this a unicode string.  should stay as binary.
        assert photo.populate('cn=foo', entry) == b'01010101'

    def test_binary_field_sanitize_for_ldap(self):
        photo = BinaryField('photo')
        # the data is already in binary representation and should remain so
        assert photo.sanitize_for_ldap(b'01010101') == b'01010101'

    def test_binary_field_pretty_print(self):
        class Foo(MyLDAPNode):
            photo = BinaryField('photo')
            name = StringField('name')

            class Meta:
                dn_format = 'ou=foos'

        obj = Foo(photo=b'aaa', name='foobar')
        obj.date_created = '2018'
        obj.date_modified = '2018'
        obj.user_created = '2018'
        obj.user_modified = '2018'

        output = obj.pretty_print()

        assert 'photo' in output
        assert 'Binary (3 bytes)' in output
        assert 'aaa' not in output
        assert 'name' in output

    def test_coerce_for_python(self):
        field = BinaryField('photo')
        assert field.coerce_for_python(b'aaa') == b'aaa'



# -*- coding: utf-8 -*-

import pytest

from ldapper.fields import StringField
from ldapper.query import Q

from .utils import MyLDAPNode


class Person(MyLDAPNode):
    uid = StringField('uid', primary=True)
    firstname = StringField('givenName', optional=True)
    lastname = StringField('sn', optional=True)


class TestQ:

    def test_one_condition(self):
        assert Q(firstname='Liam').compile(Person) == '(givenName=Liam)'

    def test_or_conditions(self):
        qset = Q(firstname='Liam') | Q(lastname='Monahan')
        assert qset.compile(Person) == '(|(givenName=Liam)(sn=Monahan))'

    def test_or_conditions_same_attr(self):
        qset = Q(firstname='Liam') | Q(firstname='Bob')
        assert qset.compile(Person) == '(|(givenName=Liam)(givenName=Bob))'

    def test_and_conditions(self):
        qset = Q(firstname='Liam') & Q(lastname='Monahan')
        assert qset.compile(Person) == '(&(givenName=Liam)(sn=Monahan))'

    def test_complex_conditions(self):
        qset = (Q(firstname='Alice') | Q(firstname='Bob')) & Q(uid='liam')
        assert qset.compile(Person) == '(&(|(givenName=Alice)(givenName=Bob))(uid=liam))'

    def test_complex_conditions2(self):
        qset = (
            (Q(firstname='Alice') | Q(firstname='Bob')) &
            (Q(uid='liam') | Q(uid='derek'))
        )
        assert qset.compile(
            Person) == '(&(|(givenName=Alice)(givenName=Bob))(|(uid=liam)(uid=derek)))'

    def test_complex_conditions3(self):
        # Test that And() & And() doesn't nest more than it needs to.
        # Same goes for Or() | Or().
        qset = (Q(lastname='Lennon') & Q(lastname='McCartney')) & (
            Q(lastname='Harrison') & Q(lastname='Starr'))
        assert qset.compile(
            Person) == '(&(sn=Lennon)(sn=McCartney)(sn=Harrison)(sn=Starr))'

        qset = (Q(lastname='Lennon') | Q(lastname='McCartney')) | (
            Q(lastname='Harrison') | Q(lastname='Starr'))
        assert qset.compile(
            Person) == '(|(sn=Lennon)(sn=McCartney)(sn=Harrison)(sn=Starr))'

    def test_complex_conditions4(self):
        qset = (Q(lastname='Lennon') & Q(lastname='McCartney')) & \
            Q(firstname='Ringo', lastname='Starr')
        assert qset.compile(
            Person) == '(&(sn=Lennon)(sn=McCartney)(givenName=Ringo)(sn=Starr))'

        qset = (Q(lastname='Lennon') | Q(lastname='McCartney')) | \
            Q(firstname='Ringo', lastname='Starr')
        assert qset.compile(
            Person) == '(|(sn=Lennon)(sn=McCartney)(&(givenName=Ringo)(sn=Starr)))'

    def test_deeply_nested(self):
        qset = (
            Q(firstname='Liam') &
            (Q(lastname='Smith') | (
                Q(uid='liam') & Q(lastname='Monahan')
            ))
        )
        assert qset.compile(
            Person) == '(&(givenName=Liam)(|(sn=Smith)(&(uid=liam)(sn=Monahan))))'

    def test_deeply_nested2(self):
        qset = (
            Q(firstname='Liam') |
            (Q(lastname='Smith') & (
                Q(uid='liam') & Q(lastname='Monahan')
            ))
        )
        assert qset.compile(Person) == \
            '(|(givenName=Liam)(&(sn=Smith)(uid=liam)(sn=Monahan)))'

        qset = Q(firstname='Liam') | (Q(lastname='Smith') | Q(uid='liam'))
        assert qset.compile(Person) == \
            '(|(givenName=Liam)(sn=Smith)(uid=liam))'

    def test_deeply_nested3(self):
        qset = (Q(uid='liam') & Q(lastname='Monahan') & Q(firstname='Liam'))
        assert qset.compile(
            Person) == '(&(uid=liam)(sn=Monahan)(givenName=Liam))'

        qset = (Q(uid='liam') | Q(lastname='Monahan') | Q(firstname='Liam'))
        assert qset.compile(
            Person) == '(|(uid=liam)(sn=Monahan)(givenName=Liam))'

        qset = (
            Q(firstname='Liam') |
            (Q(lastname='Smith') & (
                Q(uid='liam') & Q(lastname='Monahan') & Q(firstname='Liam')
            ))
        )
        assert qset.compile(Person) == \
            '(|(givenName=Liam)(&(sn=Smith)(uid=liam)(sn=Monahan)(givenName=Liam)))'

    def test_multi_condition(self):
        qset = Q(firstname='Liam', lastname='Monahan', uid='liam')
        assert qset.compile(Person) == '(&(givenName=Liam)(sn=Monahan)(uid=liam))'

        qset = Q(firstname='Liam', lastname='Monahan') | Q(uid='liam')
        assert qset.compile(Person) == '(|(&(givenName=Liam)(sn=Monahan))(uid=liam))'

    def test_attribute_misspelling(self):
        with pytest.raises(AttributeError):
            Q(missingattr='foo').compile(Person)

    def test_wrong_type(self):
        with pytest.raises(TypeError):
            Q(foo='bar') & 'foobar'
        with pytest.raises(TypeError):
            Q(foo='bar') | 'foobar'



import logging
from functools import partial
import datetime

import ldap

from ldapper.exceptions import ArgumentError
from ldapper.fields import Field, BinaryField
from ldapper.logging import ProxyLogger
from ldapper.utils import (
    bolded,
    inflect_given_cardinality,
    build_ldap_filter,
    list_items_to_sentence,
    middle_dn,
    print_word_list,
    remove_empty_strings,
    stringify,
)


log = logging.getLogger(__name__)

DEFAULT_OPTIONS = (
    'objectclasses', 'excluded_objectclasses',
    'primary_dnprefix', 'secondary_dnprefix', 'identifying_attrs',
    'searchable_fields', 'human_readable_name', 'dn_format',
)


class Options:

    def __init__(self, meta, cls_name):
        self.meta = meta
        self.objectclasses = []
        self.excluded_objectclasses = []
        self.primary_dnprefix = None
        self.secondary_dnprefix = None
        self.identifying_attrs = []
        self.searchable_fields = []
        self.human_readable_name = cls_name

        if meta:
            for attr_name in DEFAULT_OPTIONS:
                if hasattr(self.meta, attr_name):
                    setattr(self, attr_name, getattr(self.meta, attr_name))


class LDAPNodeBase(type):

    def __new__(cls, name, bases, attrs):
        new_cls = super(LDAPNodeBase, cls).__new__(cls, name, bases, attrs)

        attr_meta = attrs.pop('Meta', None)
        meta = attr_meta or getattr(new_cls, 'meta', None)

        new_cls._meta = Options(meta, cls_name=name)

        # Keep track of the fields for convenience
        fields = {}

        parents = [b for b in bases if isinstance(b, LDAPNodeBase)]

        # attrs only contains the attributes defined explicity on this
        # class.  We need to layer in the base classes' attrs.
        for parent in parents:
            fields.update(parent._fields)

        for name, attr in attrs.items():
            if isinstance(attr, Field):
                fields[name] = attr
                if attr.readonly:
                    # this is not the prettiest section of code, but hear me
                    # out: if a Field is readonly, we create it as a property.
                    # This lets us control further modification of the field.

                    # name of the concrete attr that will hold the value
                    ro_name = '__' + name

                    # we create the fget, fset, and fdel functions needed for
                    # a property.  Closures in Python are late-binding, so we
                    # use partial function application to "capture" the
                    # correct value of `name` right here and now.

                    def _fget(name, self):
                        return getattr(self, name)

                    def _fset(name, self, value):
                        self.logger.warning("Cannot modify read-only field '%s'" % name)

                    def _fdel(name, self):
                        self.logger.warning("Cannot delete read-only field '%s'" % name)

                    fget = partial(_fget, ro_name)
                    fset = partial(_fset, name)
                    fdel = partial(_fdel, name)
                    doc = '%s field' % name

                    prop = property(fget, fset, fdel, doc)

                    # set `name` to be our property
                    # set `__name` to be an initial value
                    setattr(new_cls, name, prop)
                    setattr(new_cls, ro_name, attr.default_value())
                else:
                    setattr(new_cls, name, attr.default_value())

        # at this point the parent fields and the immediate fields have both been
        # merged into fields.

        # construct the attrlist. Use to ask ldap for specific attrs to return
        new_cls._attrlist = [f.ldap for f in fields.values()]

        new_cls._system_fields = {n: f for n, f in fields.items() if f.system}
        new_cls._non_system_fields = {n: f for n, f in fields.items() if not f.system}
        new_cls._fields = fields

        # Ensure that there is never more than one primary Field
        #
        # It might be acceptable for a class to have no primary fields
        # yet.  It might be a mixin or an LDAPNode with no fields that is
        # being created to set the connection class, which other classes will
        # inherit from.
        primaries = {k: f for (k, f) in new_cls._fields.items() if f.primary is True}
        if len(primaries) > 1:
            emsg = "%s can only have at most one primary field." % new_cls.__name__
            raise ValueError(emsg)
        if primaries:
            new_cls.primary = list(primaries.keys())[0]
            new_cls._primary_field = list(primaries.values())[0]

        # The class is now ready
        return new_cls


class LDAPNode(object, metaclass=LDAPNodeBase):

    sensitive_attributes = ['userPassword']

    def __init__(self, **kwargs):
        """
        Construct a new LDAPNode.

        Raises:
            TypeError: Raised if unrecognized kwargs are passed in.
        """
        self.logger = ProxyLogger(log)
        self.conn = self.__class__.connection.get_connection()
        self._check_required_fields_present(kwargs.keys())

        for name, value in kwargs.items():
            try:
                field = self._fields[name]
            except KeyError:
                msg = "'%s' is an invalid keyword argument for this function" % name
                raise TypeError(msg)
            # readonly fields are normally not writable.  The init method
            # bypasses this restriction by directly accessing the hidden
            # field that the property is backed by.
            if field.readonly:
                setattr(self, '__' + name, field.coerce_for_python(value))
            else:
                setattr(self, name, field.coerce_for_python(value))

    def __repr__(self):
        return self.dn

    def __hash__(self):
        """
        Return the xor of the attributes that are themselves hashable.

        A meaningful implementation of this method is required to use sets.

        Lists do not define hash values.  We do not attempt to incorporate
        them into the hash value because we would have to sort the lists and
        then assume that the elements of the list are hashable.  This means
        that two objects with different values in their lists could hash to
        the same value.
        """
        hash_result = hash("LTM")
        for attr in self._non_system_fields.keys():
            try:
                hash_result = hash_result ^ hash(getattr(self, attr))
            except TypeError:
                pass
        return hash_result

    def __eq__(self, other):
        if not other:
            return False
        if self.__class__ != other.__class__:
            return False
        for attr in self._non_system_fields.keys():
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True

    def __lt__(self, other):
        return getattr(self, self.primary) < getattr(other, other.primary)

    def __str__(self):
        """
        Return a string representation of the LDAPNode.

        This requires the LDAPNode inherited class to have the dn attribute
        and the attrs list set.

        Unprintable attributes *are* printed for __str__, but not for
        pretty_print().
        """
        output = "DN: %s" % self.dn
        # first loop the attrs to figure out what the largest string is
        length = max([len(attr) for attr in self._fields.keys() if getattr(self, attr)])
        # add one more for padding
        length += 1
        output_format = '\n%%%ds: %%s' % length

        for attr in self._fields.keys():
            if isinstance(getattr(self, attr), list):
                for a in getattr(self, attr):
                    if isinstance(a, LDAPNode):
                        output += output_format % (attr, a.dnattr())
                    else:
                        output += output_format % (attr, str(a))
            elif isinstance(getattr(self, attr), LDAPNode):
                output += output_format % (attr, getattr(self, attr).dnattr())
            else:
                if getattr(self, attr) is None:
                    continue
                output += output_format % (attr, str(getattr(self, attr)))
        return output

    def pretty_print(self):
        """
        Return the prettified string version of the object

        Any Fields listed as printable=False will be omitted.
        """
        output = "DN: %s" % self.dn
        length = 0
        bolded_length = 0

        attr_names = list(self._fields.keys())

        # figure out what the longest attribute name is
        # add one more for padding
        length = max(len(attr) for attr in attr_names) + 1
        bolded_length = max(len(bolded(attr)) for attr in attr_names) + 1

        line_length = 79 - length - 1
        output_format = '\n%%%ds: %%s' % bolded_length

        # now that the max line length has been determined, we need to split
        # up the system attributes from all other attributes so that we can
        # print the system attributes last.
        system_attr_names = [n for n, f in self._fields.items() if f.system]
        attr_names = [a for a in attr_names if a not in system_attr_names]

        for attr in attr_names + system_attr_names:
            val = getattr(self, attr)
            if val is None:
                continue

            field = self._fields.get(attr, None)
            if field and not field.printable:
                continue
            if field and isinstance(field, BinaryField):
                o = 'Binary (%d bytes)' % len(val)
                output += output_format % (bolded(attr), o)
            elif isinstance(val, list):
                if len(val) > 0:
                    word_list = print_word_list(val, line_length=line_length)
                    lines = word_list.split("\n")
                    list_output_format = '\n%%%ds:%%s' % bolded_length
                    output += list_output_format % (bolded(attr), lines[0])
                    for line in lines[1:]:
                        tmp_output_format = '\n%s %%s' % (' ' * length)
                        output += tmp_output_format % (line)
            elif isinstance(val, LDAPNode):
                output += output_format % (bolded(attr), val.dnattr())
            elif isinstance(val, datetime.datetime):
                if val.tzinfo:
                    fmt = '%Y-%m-%d %H:%M:%S %Z'
                else:
                    fmt = '%Y-%m-%d %H:%M:%S'
                output += output_format % (bolded(attr), val.strftime(fmt))
            else:
                output += output_format % (bolded(attr), val)

        return output

    @property
    def hrn(self):
        """Return human-readable name."""
        return self._meta.human_readable_name

    def _dn(self):
        """
        Return the DN for the current object.

        Dynamically generate the DN for the current object based on the
        current values of the attributes involved in DN construction.
        """
        return '{},{}'.format(self._meta.dn_format % self.dnattrs(), self.conn.basedn)

    @property
    def dn(self):
        return self._dn()

    @dn.setter
    def dn(self, value):
        raise ValueError('DN cannot be set since it is dynamically generated.')

    def dnattr(self):
        """
        Return the RDN attribute value as represented in the Python object,
        not as it necessarily is in the LDAP.
        """
        # TODO this function can have a better name
        return getattr(self, self.primary)

    def dnattrs(self):
        """
        Return a hash mapping attributes to their values for attributes that
        are necessary to uniquely identify the object.  This is useful when
        constructing the DN or when calling a refetch().
        """
        attrs = {}
        for attr in self._meta.identifying_attrs:
            # for lists we will use the first value
            # TODO maybe having ListFields be an identifying_attr should
            # be an error at the time of metaclass creation.
            if isinstance(getattr(self, attr), list):
                attrs[attr] = getattr(self, attr)[0]
            else:
                attrs[attr] = getattr(self, attr)
        return attrs

    @classmethod
    def create(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    @classmethod
    def _fetch_entry(cls, primary=None, dnprefix=None, **kwargs):
        """
        Return a single result entry if unique; else None.

        Sometimes the unmanicured result of this function is all you need if
        you are trying to do a quick existence check through
        LDAPNode.obj_exists().  Otherwise, this function is used by fetch() to
        retrieve a result and parse it out into a proper object.
        """
        if kwargs:
            kwargs = stringify(kwargs)

        if primary is None and cls.primary in kwargs.keys():
            primary = kwargs[cls.primary]
        else:
            # TODO I do not think that you should be able to pass LDAPNodes in
            # if primary is an LDAPNode descendent, then get its dnattr instead
            if isinstance(primary, LDAPNode):
                primary = primary.dnattr()

        if dnprefix is None:
            if kwargs:
                try:
                    dnprefix = cls._meta.primary_dnprefix % kwargs
                except KeyError:
                    # TODO maybe this should throw an exception
                    logging.warning(
                        'Unable to fetch %s %s. Missing keyword' %
                        (cls.hrn, primary)
                        + ' argument for primary dn prefix.')
                    return None
            else:
                dnprefix = cls._meta.secondary_dnprefix
        logging.debug('Fetching %s with dnprefix %s' % (primary, dnprefix))

        conn = getattr(cls, 'connection').get_connection()
        filter = '(&(%s=%s)%s)' % \
            (cls._primary_field.ldap, primary, cls.objectclass_filter())
        basedn = '%s,%s' % (dnprefix, conn.basedn)
        try:
            result = conn.search(basedn=basedn, filter=filter, attrlist=cls._attrlist)
        except ldap.NO_SUCH_OBJECT:
            return None
        except ldap.FILTER_ERROR:
            logging.warning('Unable to fetch %s %s.  Bad LDAP search filter.' %
                            (cls.hrn, primary))
            return None

        # Return a result if and only if there is exactly one result.  If there
        # are multiple results, emit a warning, because the result of a fetch
        # should be unique.
        if result is None:
            return None
        elif len(result) == 1:
            return result.pop()
        elif len(result) == 0:
            return None
        else:
            logging.warning('Found %d results.  Unable to fetch %s %s.' %
                            (len(result), cls.hrn, primary))
            return None

    @classmethod
    def fetch(cls, primary=None, dnprefix=None, **kwargs):
        """
        Searches the LDAP for a particular object and then will attempt to
        load it into the current object.

        Arguments:
            primary -- this is the desired value of the primary identifying
                attribute to be fetched.
            dnprefix -- specify a non-conventional DN prefix.  Otherwise, the
                default one set on the class is used.
            kwargs -- a hash of the collected attributes that will be used
                     to populate the dnprefix
        """
        result = cls._fetch_entry(primary=primary, dnprefix=dnprefix, **kwargs)
        if result is not None:
            dn, entry = result
            obj = cls._parse_ldap_entry(dn, entry)
            logging.debug("Loaded %s %s Successfully" %
                          (obj.__class__.__name__, obj.dnattr()))
            return obj
        else:
            return None

    @classmethod
    def fetch_by(cls, attrs, op='&', **kwargs):
        """
        Searches the LDAP for a particular object and then will attempt to
        load it into the current object.

        Arguments:
            attrs -- a hash of the attributes to search with
            op -- operand '&', '|', '!' for the search
            kwargs -- a hash of the collected attributes that will be used
                to populate the dnprefix
        """
        if not attrs or len(attrs) == 0:
            return None
        filter = '(%s' % op
        for attr in attrs:
            filter += '(%s=%s)' % (attr, attrs[attr])
        filter += ')'
        objects = cls.list(filter=filter, max_results=1, **kwargs)
        try:
            return objects[0]
        except IndexError:
            return None

    @classmethod
    def fetch_by_dn(cls, dn, **kwargs):
        """Fetch an object when the DN is already known"""
        dn_parts = dn.split(',', 1)
        conn = getattr(cls, 'connection').get_connection()
        try:
            rdn_parts = dn_parts[0].split('=', 1)
            if len(rdn_parts) != 2:
                return None
            filter = '(%s=%s)' % tuple(rdn_parts)
        except IndexError:
            return None
        try:
            dnprefix = middle_dn(dn, ',' + conn.basedn)
        except IndexError:
            return None
        objects = cls.list(
            filter=filter, dnprefix=dnprefix, max_results=1, **kwargs)
        try:
            return objects[0]
        except IndexError:
            return None

    def refetch(self):
        """Return a fresh copy of the current object pulled from the LDAP"""
        return self.__class__.fetch(**self.dnattrs())

    def _ldap_entry(self):
        """
        Convert the current object into a dict of attributes

        You can think of this as a serialization method for objects entering
        the LDAP.
        """
        attrs = {}
        objectclasses = self.__class__._meta.objectclasses
        attrs['objectclass'] = [o.encode('utf-8') for o in objectclasses]
        for attr_name, field in self._non_system_fields.items():
            # ignore attributes starting with a "-"
            # also ignore fields that are derived (like DNPartField)
            if attr_name.startswith('-') or field.derived:
                continue

            val = getattr(self, attr_name)
            if val is not None:
                sanitized_val = field.sanitize_for_ldap(val)
                if sanitized_val:
                    attrs[field.ldap] = sanitized_val
        return attrs

    @classmethod
    def _parse_ldap_entry(cls, dn, entry):
        """Turn the raw ldap result into a Python object."""
        kwargs = {}
        for attr_name, field in cls._fields.items():
            kwargs[attr_name] = field.populate(dn, entry)
        return cls(**kwargs)

    def exists(self):
        """Return True if the current object exists in the LDAP"""
        return self.conn.exists(self.dn)

    @classmethod
    def obj_exists(cls, *args, **kwargs):
        """
        Return True if an object exists in the LDAP; False otherwise

        Pass the same arguments you would pass to fetch().
        """
        return cls._fetch_entry(*args, **kwargs) is not None

    @classmethod
    def dn_exists(cls, dn):
        """Return True if the DN exists in the LDAP"""
        return cls.connection.get_connection().exists(dn)

    def validate(self):
        """
        Validate the current LDAPNode

        This function accepts no arguments and returns True, since no
        validation criteria are known at this level.  Subclasses should
        override this method for their specific validation needs.

        Validate is called when trying to save() the current object.
        """
        return True

    def diff(self, with_ldap_names=False, skip_fake_attrs=False):
        """
        Return a hash of values that are different between self and the
        representation of the object in the LDAP.  If the object does not
        exist in the LDAP, every attribute is considered different.

        The returned hash will map a differing attribute name to a tuple
        containing (value_in_ldap, value_in_object), which is essentially
        (old_value, new_value) if self is to be saved.
        """
        results = {}
        refetch = self.refetch()
        for attr_name, field in self._non_system_fields.items():
            # derived fields do not exist as concrete fields on the entry, so
            # they should never show up in the diff.
            if field.derived:
                continue
            if skip_fake_attrs and attr_name.startswith('-'):
                continue

            self_val = getattr(self, attr_name)
            if refetch:
                refetch_val = getattr(refetch, attr_name)
            else:
                refetch_val = None

            # an attribute fetched as [] and set on the object to [''] will not
            # be equal and will be considered different, so sanitize the input
            # to avoid this sort of edge case.
            self_val = remove_empty_strings(self_val)
            refetch_val = remove_empty_strings(refetch_val)

            if self_val != refetch_val:
                if with_ldap_names:
                    results[field.ldap] = (refetch_val, self_val)
                else:
                    results[attr_name] = (refetch_val, self_val)
        return results

    def _save_existing(self):
        """Save an existing object."""
        diff = self.diff(with_ldap_names=False, skip_fake_attrs=True)
        for attr in diff:
            field = self._fields[attr]
            ldap_attr = field.ldap
            old_val, new_val = diff[attr]
            ldap_val = field.sanitize_for_ldap(new_val)
            if ldap_val and new_val is not None:
                # modifying existing attribute
                if old_val:
                    if self.conn.modify_attr(self.dn, ldap_attr, ldap_val):
                        self.log__modify_attr_success(attr, old_val, new_val)
                    else:
                        self.log__modify_attr_failure(attr, new_val)
                # adding previously nonexistant attribute
                else:
                    if self.conn.add_attr(self.dn, ldap_attr, ldap_val):
                        self.log__add_attr_success(attr, new_val)
                    else:
                        self.log__add_attr_failure(attr, new_val)
            # deleting attribute
            else:
                if self.conn.delete_attr(self.dn, ldap_attr):
                    self.log__delete_attr_success(attr)
                else:
                    self.log__delete_attr_failure(attr)

    def _save_new(self):
        """Save a new object."""
        self._before_add_callback()
        if self.conn.add(self.dn, self._ldap_entry()):
            self._after_add_callback()
            self.log__add_success()
        else:
            self.log__add_failure()

    def save(self):
        """
        Save the current LDAPNode to the LDAP

        The object MUST have its validate() method return True at the outset.
        After the object is known to be valid, save() will perform
        modifications on just the attributes that changed if the object already
        exists in the LDAP.  If it's a new object, it does an ADD instead.
        """
        # check that object is valid before saving
        if not self.validate():
            self.log__did_not_validate()
            return False

        if self.exists():
            self._save_existing()
        else:
            self._save_new()

    def delete(self):
        """Deletes the current LDAPNode in the LDAP"""
        self._before_delete_callback()
        if self.dn is not None:
            if self.conn.delete(self.dn):
                self._after_delete_callback()
                self.log__delete_success()
            else:
                self.log__delete_failure()
        else:
            self.log__delete_failure()

    @classmethod
    def filter(cls, q):
        """Return a list of LDAPNodes filter based on the given Q object(s)"""
        return cls.list(filter=q.compile(cls))

    @classmethod
    def list(cls, filter=None, prefix=None, rdn_substring=None,
             search_prefix=None, search_string=None, dnprefix=None,
             max_results=None, **kwargs):
        """
        List all objects in the directory for this class.

        Arguments:
            filter -- the caller can provide a filter to get a subset of the
                list.  This should be an LDAP search filter that INCLUDES the
                surrounding parentheses.
            prefix -- if given, return objects whose RDN starts with the
                string prefix.
            rdn_substring -- if given, return objects whose RDN contains the
                substring.
            search_prefix -- if given, return objects where any of the
                object's class's searchable_fields are matched as a prefix.
            search_string -- if given, return objects where any of the
                object's class's searchable_fields are substring matched.
            dnprefix -- Supply a DN prefix.  If given, kwargs and primary and
                secondary DN prefixes will be ignored.
            max_results -- an integer of how many results to return before
                truncation occurs.  If None, no truncation will occur.
            kwargs -- Options passed in that are used to populate the
                primary_dnprefix.

            The filter, prefix, and kwargs can all be passed and the returned
            results will have all of them satisfied.  Only the use of dnprefix
            and kwargs must be mutually exclusive.
        """
        if kwargs:
            kwargs = stringify(kwargs)

        if dnprefix is None:
            if kwargs:
                try:
                    dnprefix = cls._meta.primary_dnprefix % kwargs
                except KeyError:
                    dnprefix = cls._meta.secondary_dnprefix
            else:
                dnprefix = cls._meta.secondary_dnprefix

        logging.debug('Listing with dnprefix %s' % dnprefix)
        objs = []

        if filter is None:
            filter = cls.objectclass_filter()
        else:
            filter = '(&%s%s)' % (cls.objectclass_filter(), filter)

        if prefix:
            filter = '(&%s(%s=%s*))' % (filter, cls._primary_field.ldap, prefix)

        if rdn_substring:
            filter = '(&%s(%s=*%s*))' % (filter, cls._primary_field.ldap, rdn_substring)

        if search_prefix:
            filter = '(&%s%s)' % (
                filter,
                cls.searchable_fields_search_prefix_filter(
                    search_prefix=search_prefix))

        if search_string:
            filter = '(&%s%s)' % (
                filter,
                cls.searchable_fields_search_string_filter(
                    search_string=search_string))

        conn = cls.connection.get_connection()
        basedn = '{},{}'.format(dnprefix, conn.basedn)
        results = conn.search(basedn=basedn, filter=filter, attrlist=cls._attrlist)
        if max_results:
            num_results_processed = 0
            for dn, entry in results:
                if num_results_processed < max_results:
                    o = cls._parse_ldap_entry(dn, entry)
                    logging.debug("Loaded %s %s Successfully" %
                                  (o.__class__.__name__, o.dnattr()))
                    objs.append(o)
                    num_results_processed += 1
                else:
                    break
        else:
            for dn, entry in results:
                o = cls._parse_ldap_entry(dn, entry)
                logging.debug("Loaded %s %s Successfully" %
                              (o.__class__.__name__, o.dnattr()))
                objs.append(o)
        return objs

    @classmethod
    def _list_by(cls, op, attrs, **kwargs):
        """
        This is a helper method to return objects in the current class that
        have either the union or the intersection of the given dict of
        attributes.

        Attributes:
            op -- either '&', '|', or '!'
            attrs -- a list of tuples or a dict mapping attribute names to
                values to be matched
            kwargs -- any kwargs will be passed straight along to the
                underlying call to cls.list().
        """
        if not attrs:
            return []

        filter = '(%s' % op
        if isinstance(attrs, list):
            for k, v in attrs:
                filter += '(%s=%s)' % (k, v)
        elif isinstance(attrs, dict):
            for k, v in attrs.items():
                filter += '(%s=%s)' % (k, v)
        else:
            raise ValueError('attrs must be of type list or dict')
        filter += ')'

        return cls.list(filter=filter, **kwargs)

    @classmethod
    def list_by_union(cls, attrs, **kwargs):
        """
        Return a list of LDAPNodes representing the union of the given
        attribute values

        Attributes:
            attrs -- a list of tuples or a dict mapping attribute names to
                values to be matched
        """
        return cls._list_by(op='|', attrs=attrs, **kwargs)

    @classmethod
    def list_by_negation(cls, attrs, **kwargs):
        """
        Return a list of LDAPNodes represented by the negation of the
        given attritbute and value.  There can only be one attribute
        value pair given, otherwise a exception will be raised.

        Attributes:
            attrs -- a list of tuples or a dict mapping attribute names to
                values to be matched
        """
        if len(attrs) != 1:
            raise ArgumentError('Only one attribute value pair supported')
        return cls._list_by(op='!', attrs=attrs, **kwargs)

    @classmethod
    def list_by(cls, attrs, **kwargs):
        """
        Return a list of LDAPNodes representing the intersection of the
        given attribute values

        Attributes:
            attrs -- a list of tuples or a dict mapping attribute names to
                values to be matched
        """
        return cls._list_by(op='&', attrs=attrs, **kwargs)

    @classmethod
    def _searchable_fields_filter(cls, search_term, template,
                                  searchable_fields=None):
        """
        A helper method for searchable fields

        template must have two value placeholders for the var and the value
        """
        filter = ''

        if searchable_fields is None:
            searchable_fields = cls._meta.searchable_fields
        # TODO searchable_fields, if passed in, will need to be converted from
        # Python names to LDAP names

        for field in searchable_fields:
            filter += template % (field, search_term)

        return '(|%s)' % filter

    @classmethod
    def searchable_fields_search_prefix_filter(cls, search_prefix,
                                               searchable_fields=None):
        """Return a search prefix filter using the searchable_fields"""
        return cls._searchable_fields_filter(
            search_term=search_prefix, template='(%s=%s*)',
            searchable_fields=searchable_fields)

    @classmethod
    def searchable_fields_search_string_filter(cls, search_string,
                                               searchable_fields=None):
        """Return a search substring filter using the searchable_fields"""
        templates = ['(%s=%s)', '(%s=%s*)', '(%s=*%s)', '(%s=*%s*)']
        filter = ''
        for template in templates:
            filter = filter + cls._searchable_fields_filter(
                search_term=search_string, template=template,
                searchable_fields=searchable_fields)
        return '(|%s)' % filter

    def has_attrval(self, var, value):
        """Return True if the var has the value in question; False otherwise"""
        return value in getattr(self, var)

    def _obscure_if_sensitive(self, attr_name, value):
        """
        Return "*****" if the attribute is sensitive.  If the attribute is
        not sensitive, return the actual value.

        The attr_name is the name in LDAP.
        """
        if attr_name in self.__class__.sensitive_attributes:
            return '*****'
        return value

    # Callbacks/Hooks
    # Child classes can override these methods to do something useful
    def _before_add_callback(self):
        """Gets called just before the LDAPNode is added to the LDAP."""
        pass

    def _after_add_callback(self):
        """Gets called just after the LDAPNode is added to the LDAP."""
        pass

    def _before_delete_callback(self):
        """Gets called just before the LDAPNode is deleted from the LDAP."""
        pass

    def _after_delete_callback(self):
        """Gets called just after the LDAPNode is deleted from the LDAP."""
        pass

    # Log messages
    def log__did_not_validate(self):
        """Log a faiure to pass validation"""
        self.logger.error(
            "%s %s did not pass validation." % (self.hrn, self.dnattr()))

    def log__add_success(self):
        """Log a successful add of the current LDAPNode"""
        self.logger.info(
            "%s %s was added." % (self.hrn, self.dnattr()))

    def log__add_failure(self):
        """Log an unsuccessful add of the current LDAPNode"""
        self.logger.error(
            "Could not add %s %s." % (self.hrn, self.dnattr()))

    def log__add_attr_success(self, attr_name, val):
        """Log a successful attribute add on the current LDAPNode"""
        if isinstance(val, list) and len(val) == 1:
            val = list_items_to_sentence(val)
        self.logger.info('Successfully added %s %s to %s %s' %
                         (attr_name,
                          self._obscure_if_sensitive(attr_name, val),
                          self.hrn, self.dnattr()))

    def log__add_attr_failure(self, attr_name, val):
        """Log an unsuccessful attribute add on the current LDAPNode"""
        self.logger.info('Failed to add %s %s to %s %s' %
                         (attr_name,
                          self._obscure_if_sensitive(attr_name, val),
                          self.hrn, self.dnattr()))

    def log__modify_attr_success(self, attr_name, old_val, new_val):
        """Log a successful modification of a field on the current LDAPNode"""
        if isinstance(old_val, list) and isinstance(new_val, list):
            old_val = set(old_val)
            new_val = set(new_val)
            added = list(new_val - old_val)
            removed = list(old_val - new_val)
            if added:
                name = inflect_given_cardinality(attr_name, len(added))
                added = list_items_to_sentence(added)
                self.logger.info(
                    "Added %s %s to %s %s" %
                    (name, added, self.hrn, self.dnattr()))
            if removed:
                name = inflect_given_cardinality(attr_name, len(removed))
                removed = list_items_to_sentence(removed)
                self.logger.info(
                    "Removed %s %s from %s %s" %
                    (name, removed, self.hrn, self.dnattr()))
        else:
            self.logger.info(
                "Changed %s from %s to %s on %s %s" %
                (attr_name,
                 self._obscure_if_sensitive(attr_name, old_val),
                 self._obscure_if_sensitive(attr_name, new_val),
                 self.hrn, self.dnattr()))

    def log__modify_attr_failure(self, attr_name, attr_value):
        """Log a failed modification of a field on the current LDAPNode"""
        if isinstance(attr_value, list):
            attr_value = '"%s"' % (', '.join(attr_value))
        self.logger.error(
            "Failed to set %s to %s on %s %s" %
            (attr_name,
             self._obscure_if_sensitive(attr_name, attr_value),
             self.hrn, self.dnattr()))

    def log__delete_success(self):
        """Log a successful removal of an LDAPNode from the LDAP"""
        self.logger.info(
            "%s %s was removed successfully." %
            (self.hrn, self.dnattr()))

    def log__delete_failure(self):
        """Log an unsuccessful removal of an LDAPNode from the LDAP"""
        self.logger.error(
            "%s %s was not removed successfully." %
            (self.hrn, self.dnattr()))

    def log__delete_attr_success(self, attr_name):
        """Log a successful deletion of a field on the current LDAPNode"""
        self.logger.info('Successfully removed %s attribute from %s %s' %
                         (attr_name, self.hrn, self.dnattr()))

    def log__delete_attr_failure(self, attr_name):
        """Log an unsuccessful deletion of a field on the current LDAPNode"""
        self.logger.error('Failed to remove %s attribute from %s %s' %
                          (attr_name, self.hrn, self.dnattr()))

    @classmethod
    def objectclass_filter(cls):
        """Return a string LDAP filter from the objectClass attributes"""
        f = build_ldap_filter(
            op='&', attrname='objectClass', items=cls._meta.objectclasses)
        if cls._meta.excluded_objectclasses:
            n = build_ldap_filter(
                op='!', attrname='objectClass',
                items=cls._meta.excluded_objectclasses)
            f = '(&%s%s)' % (f, n)
        return f

    def attr_difference_since_last_save(self, attr):
        """
        For a given attribute, return a dict of the values added and removed
        between what is in the LDAP and what is in the current object.

        The `attr` in question *must* be a list for calling this method to make
        any sense.
        """
        diff = self.diff()

        if attr not in self._fields.keys():
            raise ArgumentError("%s is not a valid attribute" % attr)

        if attr not in diff:
            return {}

        old_vals, new_vals = diff[attr]
        # we need to check if either old_vals or new_vals is None, since what
        # is passed to set() must be iterable
        if old_vals is None:
            old_vals = []
        if new_vals is None:
            new_vals = []
        old_vals = set(old_vals)
        new_vals = set(new_vals)

        changeset = {}
        changeset['added'] = list(new_vals.difference(old_vals))
        changeset['removed'] = list(old_vals.difference(new_vals))
        return changeset

    def attr_added_since_last_save(self, attr):
        """
        For a given attribute, return a list of the values added between what
        is in the LDAP and what is in the current object.
        """
        return self.attr_difference_since_last_save(attr).get('added', [])

    def attr_removed_since_last_save(self, attr):
        """
        For a given attribute, return a list of the values removed between what
        is in the LDAP and what is in the current object.
        """
        return self.attr_difference_since_last_save(attr).get('removed', [])

    def _check_required_fields_present(self, attrs):
        necessary = set([f for f, val in self._fields.items() if not val.optional])
        available = set(attrs)
        missing = necessary - available
        if missing:
            if len(missing) == 1:
                msg = "Required field '{}' is missing".format(missing.pop())
            else:
                fields = ["'{}'".format(f) for f in missing]
                msg = 'Required fields [{}] are missing'.format(', '.join(fields))
            raise Exception(msg)



class LdapperError(Exception):

    """Base class for exceptions in this module.

        A msg MUST be provided.
    """

    def __str__(self):
        return '(%s) %s' % (self.__class__.__name__, self.msg)


class AddDNFailed(LdapperError):

    """Exception raised when we failed to add a DN to the LDAP

    Attributes:
       dn -- DN that failed to be added
       msg -- explanation of the error
    """

    def __init__(self, dn):
        self.dn = dn
        self.msg = 'Unable to add the DN %s to LDAP' % dn


class ArgumentError(LdapperError):

    """Exception raised when the arguments to a function are invalid

    Attributes:
        msg -- explanation of the error
    """

    def __init__(self, msg):
        self.msg = msg


class DuplicateValue(LdapperError):

    """Tried to write a duplicate value to the LDAP

    Attributes:
       attr -- Attribute name
       value -- Attribute value
    """

    def __init__(self, attr, value):
        self.attr = attr
        self.original_value = value
        if isinstance(value, list):
            dups = [x for x in value if value.count(x) >= 2]
            self.offending_values = list(set(dups))
        else:
            # I'm not sure when something other than a list could have a
            # duplicate value, but just in case...
            self.offending_values = value
        self.msg = 'Attribute "%s" has duplicate value(s): %s' % (
            attr, self.offending_values)


class NoSuchAttrValue(LdapperError):

    """Exception raised when a DN does not have a given value.

    Attributes:
        dn -- the DN that did not exist
        attribute -- the attribute that is not present
        value -- the value that the attribute was expected to have
        msg -- explanation of the error
    """

    def __init__(self, dn, attribute, value):
        self.dn = dn
        self.attribute = attribute
        self.value = value
        self.msg = 'DN %s does not have %s %s' % (dn, attribute, value)


class NoSuchDN(LdapperError):

    """Exception raised when a DN does not exist.

    Attributes:
        dn -- the DN that did not exist
        msg -- explanation of the error
    """

    def __init__(self, dn):
        self.dn = dn
        self.msg = 'DN %s does not exist.' % dn


class InvalidDN(LdapperError):

    """Exception raised when a DN does not meet RFC 4514 syntax

    Attributes:
       dn -- the invalid dn
       msg -- explanation of the error
    """

    def __init__(self, obj, name, dn):
        self.obj = obj
        self.name = name
        self.dn = dn
        if self.obj and self.name:
            self.msg = '%s %s has an invalid dn format: %s' % (obj, name, dn)
        else:
            self.msg = 'Invalid DN format: %s' % dn



#!/usr/bin/env python

from ldapper import __version__

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name="ldapper",
    version=__version__,
    description="LDAP ORM for Python",
    author="Liam Monahan",
    author_email="liam@liammonahan.com",
    url="https://github.com/UMIACS/ldapper",
    license="LGPL v2.1",
    packages=["ldapper"],
    platforms="UNIX/Linux",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Database",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP",
    ],
    install_requires=["inflection", "python-ldap>=2.4.15"],
)



# -*- coding: utf-8 -*-

from ldapper.exceptions import (
    AddDNFailed,
    ArgumentError,
    DuplicateValue,
    NoSuchAttrValue,
    NoSuchDN,
)


class TestExceptions:

    def test_AddDNFailed(self):
        error = AddDNFailed('cn=foo,ou=bar')
        assert error.dn == 'cn=foo,ou=bar'
        assert error.msg == 'Unable to add the DN cn=foo,ou=bar to LDAP'
        assert str(error) == '(AddDNFailed) Unable to add the DN cn=foo,ou=bar to LDAP'

    def test_ArgumentError(self):
        assert ArgumentError('foo bar').msg == 'foo bar'

    def test_DuplicateValue(self):
        lst = ['wine', 'cheese', 'wine']
        error = DuplicateValue(attr='foods', value=lst)
        assert error.offending_values == ['wine']
        assert error.msg == 'Attribute "foods" has duplicate value(s): [\'wine\']'

        # maybe we already know what the duplicate values are
        error = DuplicateValue(attr='foods', value='cheese')
        assert error.msg == 'Attribute "foods" has duplicate value(s): cheese'

    def test_NoSuchAttrValue(self):
        error = NoSuchAttrValue(dn='cn=foo', attribute='dest', value='Milan')
        assert error.msg == 'DN cn=foo does not have dest Milan'

    def test_NoSuchDN(self):
        error = NoSuchDN(dn='cn=foo')
        assert error.msg == 'DN cn=foo does not exist.'



import re
import logging
from getpass import getpass

import ldap
import ldap.modlist as modlist

from .exceptions import (
    AddDNFailed,
    NoSuchDN,
    NoSuchAttrValue,
    DuplicateValue,
)

log = logging.getLogger(__name__)


class BaseConnection(object):

    """
    A BaseConnection object can be used to connect to a given LDAP server.

    These settings should be set on subclasses to know the server uri and the
    basedn of the LDAP tree:

      :URI:
        The uri should contain the schema, host, and an optional port
        field.

        e.g. ldaps://ldap.example.com

      :BASEDN:
        The base DN of the LDAP tree.
    """

    __human_readable_name__ = 'LDAP'
    attrlist = [
        '*', 'createTimestamp', 'modifyTimestamp',
        'creatorsName', 'modifiersName',
    ]

    # subclasses must define: BASE_DN, and URI

    def __init__(self, logindn, password, uri=None, certfile=None,
                 basedn=None):
        if basedn:
            self.basedn = basedn
        else:
            self.basedn = self.__class__.BASE_DN

        if uri:
            self.uri = uri
        else:
            self.uri = self.URI

        if certfile:
            ldap.set_option(ldap.OPT_X_TLS_CACERTFILE, certfile)

        self.ldaps = ldap.initialize(self.uri)
        self.ldaps.set_option(ldap.OPT_PROTOCOL_VERSION, 3)

        if logindn is None and password is None:
            self.ldaps.simple_bind_s()
        else:
            fq_logindn = self._fully_qualify_dn(logindn)
            self.ldaps.simple_bind_s(fq_logindn, password)

        self.dn = self.whoami()
        log.debug('login successful: %s on %s', self.dn, self.uri)

    def _fully_qualify_dn(self, logindn):
        """Return a full-qualified login dn."""
        if not re.search('=', logindn) and not re.search('@', logindn):
            logindn = "uid=%s,ou=people,%s" % (logindn, self.basedn)
        return logindn

    @classmethod
    def connect(cls, logindn=None, password=None, uri=None,
                certfile=None, basedn=None, retries=3):
        """
        Attempt to establish a connection using the supplied parameters.

        Return a connection object.  Return None if unsuccessful.
        """
        name = cls.__human_readable_name__

        # tmpuri is the uri name that we'll print in log messages
        if uri is None:
            tmpuri = cls.URI
        else:
            tmpuri = uri

        for x in range(1, retries + 1):
            if logindn is None:
                logindn = input('Enter a %s loginDN or username: ' % name)
            if password is None:
                password = getpass('Password for %s (%s): ' % (name, logindn))
            try:
                return cls(
                    logindn=logindn,
                    password=password,
                    uri=uri,
                    certfile=certfile,
                    basedn=basedn,
                )
            except ldap.LDAPError:
                log.warning('login attempt %d failed: %s on %s', x, logindn, tmpuri)
                password = None  # prompt again for password on next try

        return None  # all retries failed

    @classmethod
    def prompt(cls, logindn=None, password=None, uri=None,
               certfile=None, basedn=None, retries=3):
        """
        Attempt to establish a connection using the supplied parameters.

        This method will set the connection if successful.

        Return the password used if successful and None if unsuccessful.
        """
        name = cls.__human_readable_name__

        # tmpuri is the uri name that we'll print in log messages
        if uri is None:
            tmpuri = cls.URI
        else:
            tmpuri = uri

        for x in range(retries):
            if logindn is None:
                logindn = input('Enter a %s loginDN or username: ' % name)
            if password is None:
                password = getpass('Password for %s (%s): ' % (name, logindn))
            try:
                conn = cls(
                    logindn=logindn,
                    password=password,
                    uri=uri,
                    certfile=certfile,
                    basedn=basedn,
                )
                log.debug('login successful: %s on %s', logindn, tmpuri)
                cls.set_connection(conn)
                return password
            except ldap.LDAPError:
                log.error('login attempt %d failed: %s on %s', x, logindn, tmpuri)
                password = None  # prompt again for password on next try

        return None  # all retries failed

    @classmethod
    def connect_anon(cls, uri=None):
        """Return an anonymous connection to the LDAP"""
        return cls(logindn=None, password=None, uri=uri)

    @classmethod
    def set_connection(cls, conn):
        cls.conn = conn

    @classmethod
    def get_connection(cls):
        """
        Return the ldap connection that was set

        If the connection object has not been set when this method is
        called, then create and set an anonymous connection object.
        """
        try:
            return cls.conn
        except AttributeError:
            cls.set_connection(cls.connect_anon())
            return cls.conn

    def __str__(self):
        return self.uri + ' as ' + self.whoami()

    def __unicode__(self):
        return self.uri + ' as ' + self.whoami()

    def __del__(self):
        self.ldaps.unbind_s()

    def is_anonymous(self):
        return self.dn is None

    def whoami(self):
        """
        Return the full DN of the entity that is bound to this connection

        Return None if this is an anonymous bind.
        """
        answer = self.ldaps.whoami_s()
        if answer is None or answer == '':
            return None  # anonymous bind

        if re.search('^dn:', answer):
            return answer.replace('dn:', '', 1)
        if re.search('^u:', answer):
            return answer.replace('u:', '', 1)

        return None  # should be virtually unreachable

    def whoami_short(self):
        dn = self.whoami()
        if dn is None:
            return dn
        parts = dn.split(',')
        return parts[0].split('=')[1]

    def exists(self, dn):
        scope = ldap.SCOPE_BASE
        try:
            ldap.dn.explode_dn(dn)
        except ldap.DECODING_ERROR:
            return False
        try:
            self.ldaps.search_s(dn, scope)
            return True
        except ldap.NO_SUCH_OBJECT:
            return False

    def search(self, basedn=None, scope=ldap.SCOPE_SUBTREE,
               filter='(objectClass=*)', attrlist=None):
        """
        Perform an LDAP search operation.

        Optional attributes are a basedn, a scope, and a filter.  Use
        ldap.SCOPE_BASE to search the object represented by the basedn itself.
        The default search scope is ldap.SCOPE_SUBTREE.
        """
        if basedn is None:
            basedn = self.basedn
        if attrlist is None:
            attrlist = self.__class__.attrlist
        log.debug('Searching with filter %s in %s under scope %s' %
                  (filter, basedn, scope))
        try:
            return self.ldaps.search_s(basedn, scope, filter, attrlist=attrlist)
        except ldap.SERVER_DOWN:
            log.error('Could not contact the LDAP service.  Server down.')
            # sometimes connections go stale if the server has restarted.

            # there is nothing we can do for un-anonymous connections.
            if not self.is_anonymous():
                raise

            # if the connection is anonymous, we will retry the search
            # using a fresh connection.
            log.warning('Recreating a new anonymous connection.')

            cls = self.__class__
            new_conn = cls.connect_anon()

            # if this search still fails we won't catch it.  If we get
            # past that point, then set the new anon connection and
            # return the result
            #
            # Nota bene: this code is intentionally not recursive!
            result = new_conn.ldaps.search_s(basedn, scope, filter,
                                             attrlist=attrlist)

            cls.set_connection(new_conn)
            return result

    def add(self, dn, attrs):
        try:
            self.ldaps.add_s(dn, modlist.addModlist(attrs))
            log.debug('add %s' % dn)
            return True
        except ldap.INVALID_DN_SYNTAX:
            raise
        except ldap.NO_SUCH_OBJECT:
            raise AddDNFailed(dn)
        except ldap.LDAPError:
            log.error('add %s failed' % dn, exc_info=False)
            raise

    def delete(self, dn):
        try:
            self.ldaps.delete_s(dn)
            log.debug('delete %s' % dn)
            return True
        except ldap.LDAPError:
            log.error('delete %s failed' % dn, exc_info=False)
            return False

    def modify_attr(self, dn, attribute, value):
        tuple = [(ldap.MOD_REPLACE, attribute, value)]
        try:
            self.ldaps.modify_s(dn, tuple)
            log.debug('mod attr %s on %s' % (attribute, dn))
            return True
        except ldap.NO_SUCH_OBJECT:
            raise NoSuchDN(dn=dn)
        except ldap.TYPE_OR_VALUE_EXISTS:
            raise DuplicateValue(attr=attribute, value=value)
        except ldap.LDAPError:
            log.error('mod attr %s on %s failed' % (attribute, dn),
                      exc_info=False)
            raise
            return False

    def add_attr(self, dn, attribute, value):
        tuple = [(ldap.MOD_ADD, attribute, value)]
        try:
            self.ldaps.modify_s(dn, tuple)
            log.debug('add attr %s on %s' % (attribute, dn))
            return True
        except ldap.NO_SUCH_OBJECT:
            raise NoSuchDN(dn=dn)
        except ldap.TYPE_OR_VALUE_EXISTS:
            raise DuplicateValue(attr=attribute, value=value)
        except ldap.LDAPError:
            log.error('add attr %s on %s failed' % (attribute, dn),
                      exc_info=False)
            raise
            return False

    def delete_attr(self, dn, attribute, value=None):
        tuple = [(ldap.MOD_DELETE, attribute, value)]
        try:
            self.ldaps.modify_s(dn, tuple)
            log.debug('del attr %s and value %s on %s' %
                      (attribute, dn, value))
            return True
        except ldap.NO_SUCH_OBJECT:
            raise NoSuchDN(dn=dn)
        except ldap.NO_SUCH_ATTRIBUTE:
            raise NoSuchAttrValue(dn=dn, attribute=attribute, value=value)
        except ldap.LDAPError:
            log.error('del attr %s and value %s on %s failed' %
                      (attribute, value, dn), exc_info=False)
            raise
            return False

    def ldif(self, filter):
        """Return an LDIF string of all results returned for the given filter"""
        log.debug("Using search filter : %s" % filter)

        results = self.search(filter=filter)
        if len(results) == 1:
            log.info("Found 1 entry")
        else:
            log.info("Found %s entries" % len(results))

        output = ''
        for dn, entry in results:
            output += "-" * 72 + "\n"
            output += "DN: %s\n" % dn
            length = 0
            # first loop the attrs to figure out what the largest string is
            for attr in entry.keys():
                if len(attr) > length:
                    length = len(attr)
            # add one more for padding
            length += 1
            output_format = '%%%ds: %%s\n' % length
            for attr, values in entry.items():
                for val in values:
                    output += output_format % (attr, val.decode())
            output += "\n"
        return output



# -*- coding: utf-8 -*-

import pytest

from datetime import datetime

from ldapper.connection import BaseConnection
from ldapper.ldapnode import LDAPNode
from ldapper.fields import (
    ListField,
    StringField,
    SystemField,
)


class SystemStringField(SystemField, StringField):
    pass


class SystemDateField(SystemField, StringField):

    def coerce_for_python(self, value):
        if value:
            return datetime.strptime(value, "%Y%m%d%H%M%SZ")


class connection(BaseConnection):
    BASE_DN = 'dc=acme,dc=org'
    URI = 'ldap://localhost:389'


class MyLDAPNode(LDAPNode):
    connection = connection

    date_created = SystemDateField('createTimestamp', optional=True)
    date_modified = SystemDateField('modifyTimestamp', optional=True)
    user_created = SystemStringField('creatorsName', optional=True)
    user_modified = SystemStringField('modifiersName', optional=True)


class Person(MyLDAPNode):
    uid = StringField('uid', primary=True)
    lastname = StringField('sn')
    fullname = StringField('cn')
    addresses = ListField('mailLocalAddress', optional=True)

    class Meta:
        objectclasses = ['top', 'inetOrgPerson']
        dn_format = 'uid=%(uid)s,ou=people'
        primary_dnprefix = 'ou=people'
        secondary_dnprefix = 'ou=people'
        identifying_attrs = ['uid']
        searchable_fields = ['uid', 'cn', 'displayName', 'givenName', 'sn']
        human_readable_name = 'LDAP Person'


person_kwargs = {'uid': 'liam', 'lastname': 'Monahan', 'fullname': 'Liam Monahan'}


# TODO this should get moved somewhere more central
conn = connection.connect(logindn='cn=admin,dc=acme,dc=org',
                          password='JonSn0w')
connection.set_connection(conn)


# TODO this should also get moved somewhere else once we get further along
def get_person():
    return Person(**person_kwargs)


class TestLDAPNode:

    def test_default_meta_options(self):
        class Test(MyLDAPNode):
            pass
        assert Test._meta.objectclasses == []
        assert Test._meta.excluded_objectclasses == []
        assert Test._meta.primary_dnprefix is None
        assert Test._meta.secondary_dnprefix is None
        assert Test._meta.identifying_attrs == []
        assert Test._meta.searchable_fields == []
        assert Test._meta.human_readable_name == 'Test'

    def test_ldapnode_metaclass_new(self):
        """Test that the __new__ functionality of the metaclass is working."""
        assert Person._meta.human_readable_name == 'LDAP Person'
        assert Person.primary == 'uid'
        assert type(Person._fields['fullname']) is StringField

        # test that fields are inheritable
        class SpecialPerson(Person):
            superpower = StringField('superpower')
            lastname = StringField('sn', optional=True)
        assert 'superpower' in SpecialPerson._fields
        assert 'uid' in SpecialPerson._fields

        # lastname should be overidden in the child class
        assert SpecialPerson._fields['lastname'].optional

    def test_readonly_field(self, caplog):
        class ROPerson(Person):
            uid = StringField('uid', primary=True, readonly=True)
        person = ROPerson(uid='liam', lastname='Alpert', fullname='Ram Dass')

        # cannot set
        person.uid = 'different'
        assert person.uid == 'liam'
        assert "Cannot modify read-only field 'uid'" in caplog.text

        # cannot delete
        del person.uid
        assert person.uid == 'liam'
        assert "Cannot delete read-only field 'uid'" in caplog.text

    def test_ldapnode_required_fields_present(self):
        with pytest.raises(Exception) as excinfo:
            Person()
        assert 'are missing' in str(excinfo.value)

    def test_invalid_kwargs_to_ldapnode_init(self):
        with pytest.raises(TypeError):
            Person(uid='a', lastname='l', fullname='f', invalid='foo')

    def test_primary_field_is_inherited(self):
        # the primary field needs to be inherited
        class UnixPerson(Person):
            pass
        assert UnixPerson.primary == 'uid'

    def test_too_many_primary_fields(self):
        with pytest.raises(ValueError):
            class BogusPerson(Person):
                another_primary = StringField('anotherPrimary', primary=True)

    def test_non_system_fields(self):
        assert (
            sorted(['addresses', 'fullname', 'lastname', 'uid']) ==
            sorted(Person._non_system_fields.keys())
        )
        assert (
            sorted(['date_created', 'date_modified', 'user_created', 'user_modified']) ==
            sorted(Person._system_fields.keys())
        )

    def test_system_fields(self):
        person = Person(**person_kwargs)
        person.save()

        # test that system fields are not included in the diff
        assert person.diff() == {}

        # a new object does not have a date_created yet
        # you must refetch to populate those values
        assert person.date_created is None
        person = person.refetch()
        assert person.date_created is not None

        # changing the system fields should not have any effect since
        # they will never be saved back to ldap
        person.date_created = datetime.now()
        assert person.diff() == {}

        # system fields shall not used to generate the object hash
        hash_before = hash(person)
        person.date_modified = 'foo'
        assert hash_before == hash(person)

        person.delete()

    def test_ldapnode_repr(self):
        p = Person(**person_kwargs)
        assert repr(p) == 'uid=liam,ou=people,dc=acme,dc=org'

    def test_ldapnode_hash(self):
        p = Person(**person_kwargs)
        assert hash(p) == hash(p)

    def test_ldapnode_eq(self):
        p1 = Person(**person_kwargs)
        p2 = Person(**person_kwargs)
        assert not p1.__eq__(None)
        assert not p1.__eq__(True)
        assert not p1.__eq__(False)

        assert p1.__eq__(p2)
        p2.lastname = 'Caesar'
        assert not p1.__eq__(p2)

    def test_ldapnode_ordering(self):
        aaa = Person(uid='aaa', lastname='ln', fullname='fn')
        bbb = Person(uid='bbb', lastname='ln', fullname='fn')
        lst = [bbb, aaa]
        lst.sort()  # will sort by the primary attr: uid in this case
        assert lst[0] == aaa
        assert lst[1] == bbb

    def test_ldapnode_str(self):
        # this test should test differently user python 2 and python 3
        # this test is making sure that bost ListFields and other fields
        # use the dnattr when there is an LDAPNode slotted in there.
        p = Person(
            uid='liam',
            lastname=Person(**person_kwargs),
            fullname=None,
            addresses=['liam@acme.org', Person(**person_kwargs)]
        )
        expected = """DN: uid=liam,ou=people,dc=acme,dc=org
       uid: liam
  lastname: liam
 addresses: liam@acme.org
 addresses: liam"""
        assert expected == str(p)

    def test_ldapnode_dn(self):
        p = Person(**person_kwargs)

        # test that dn() works
        assert p.dn == 'uid=liam,ou=people,dc=acme,dc=org'

        # setting the dn should raise an error
        with pytest.raises(ValueError):
            p.dn = 'uid=foo'

        # test dnattr()
        assert p.dnattr() == 'liam'

        # test dnattrs()
        assert p.dnattrs() == {'uid': 'liam'}

    def test_ldapnode_create(self):
        person = Person.create(**person_kwargs)
        assert person.uid == 'liam'

    def test_fetch(self):
        p = Person(**person_kwargs)
        p.save()

        assert Person.fetch('liam') == p
        assert Person.fetch(uid='liam') == p
        assert Person.fetch(p) == p

        assert Person.fetch('who') is None

        p.delete()

    def test_fetch_by(self):
        p = Person(**person_kwargs)
        p.save()

        assert Person.fetch_by(attrs={}) is None
        assert Person.fetch_by(attrs={'uid': p.uid}) == p
        assert Person.fetch_by(attrs={'uid': p.uid + 'a'}) is None

        p.delete()

    def test_fetch_by_dn(self):
        p = Person(**person_kwargs)
        p.save()

        assert Person.fetch_by_dn('uid=kfjsdlkjf,ou=people,dc=acme,dc=org') is None
        assert Person.fetch_by_dn('uid,ou=people,dc=acme,dc=org') is None
        assert Person.fetch_by_dn('malformed') is None
        assert Person.fetch_by_dn(p.dn) == p

        # for setUp/tearDown efficiency's sake, test obj_exists() here too
        assert Person.obj_exists(p.uid)
        assert not Person.obj_exists(p.uid + 'a')

        assert Person.dn_exists(p.dn)

        p.delete()

    def test_refetch(self):
        p = Person(**person_kwargs)
        p.save()
        p.lastname = p.lastname + 'aaa'

        untainted_person = Person.fetch(p.uid)
        assert untainted_person != p
        assert untainted_person == p.refetch()

        # assert that refetch does not modify the existing object
        assert untainted_person != p

        p.delete()

    def test_diff(self):
        person = Person(
            uid='liam',
            lastname='Monahan',
            fullname='Liam Monahan',
        )

        assert person.diff() == {
            'addresses': (None, []),
            'fullname': (None, 'Liam Monahan'),
            'lastname': (None, 'Monahan'),
            'uid': (None, 'liam')
        }

        person.save()

        assert person.diff() == {}

        person.lastname = 'Jones'
        assert person.diff() == {'lastname': ('Monahan', 'Jones')}
        assert person.diff(with_ldap_names=True) == {'sn': ('Monahan', 'Jones')}

        person.delete()

    def test_happy_path_crud(self):
        person = Person(
            uid='liam',
            lastname='Monahan',
            fullname='Liam Monahan',
        )
        person.save()
        assert person.exists()

        person.delete()
        assert not person.exists()
